import pandas as pd
import os
from tqdm.auto import tqdm
from typing import List
from multiprocessing import Pool, cpu_count
from .extract import load_epigenomes_table
from .mine import get_target_path


def to_dict(path: str, target: str):
    df = pd.read_csv(path, sep="\t", index_col=[0, 1, 2, 3]).round(2)
    df.columns = pd.MultiIndex.from_product([[target], df.columns])
    return df


def _to_dict(kwargs):
    return to_dict(**kwargs)


def concatenate(
    root: str,
    cell_lines: List[str],
    assembly: str,
    workers: int
):
    """Concatenate the targets into a single file.

    Parameters
    ----------------------------------
    root: str,
        Root from where to search the files
        to concatenate.
    cell_lines: List[str],
        Cell lines to consider.
    assembly: str,
        The genomic assembly of the data to be retrieved.
    workers: int,
        Workers to use to parallelize the loading.
    """
    table = load_epigenomes_table(cell_lines, assembly)
    if workers == -1:
        workers = cpu_count()
    with Pool(min(cpu_count(), workers)) as p:
        for cell_line, group in tqdm(table.groupby("cell_line"), leave=False, desc="Concatenating cell lines"):
            path = "{root}/{assembly}/{cell_line}.csv.xz".format(
                root=root,
                assembly=assembly,
                cell_line=cell_line
            )
            if os.path.exists(path):
                continue
            paths = [
                {
                    "path": get_target_path(root, cell_line, assembly, target),
                    "target": target
                }
                for target in group.target.unique()
            ]

            pd.concat(list(tqdm(
                p.imap(_to_dict, paths),
                desc="Concatenating files",
                total=len(paths),
                leave=False
            )), axis=1).reset_index().to_csv(path, index=False)

        p.close()
        p.join()



from typing import Tuple
import os
from encodeproject.utils import download
import pandas as pd


def load_epigenomes(
    cell_line: str = "K562",
    assembly: str = "hg38",
    dataset: str = "fantom",
    region: str = "promoters",
    window_size: int = 256,
    root: str = "datasets",
    drop_unique_group_by: bool = True
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Return epigenomic data and labels for given parameters.

    Parameters
    ----------------------------------------
    cell_line: str = "K562",
        Cell line to consider. By default K562.
        Currently available cell lines are
        listed in the repository README file.
    assembly: str,
        The genomic assembly of the data to be retrieved.
    dataset: str = "fantom",
        Dataset to consider. By default fantom.
        Currently available datasets are
        listed in the repository README file.
    region: str = "promoters",
        Region to consider. By default promoters.
        Currently available region are
        listed in the repository README file.
    window_size: int = 200,
        Window size to consider. By default 200.
        Currently available window sizes are
        listed in the repository README file.
    root: str = "datasets"
        Where to store the downloaded data.
    drop_unique_group_by: bool = True,
        Whetever to drop the group by column layer
        when it is unique, eg when only a max group by
        is available in the dataset.

    Returns
    ----------------------------------------
    Return tuple with input and output DataFrames.
    """
    repository = "https://github.com/LucaCappelletti94/epigenomic_dataset/blob/master/preprocessed"
    get_parameter = "?raw=true"
    data_path_placeholder = "{{root}}/{dataset}/{assembly}/{window_size}/{region}/{cell_line}.csv.xz".format(
        root=root,
        dataset=dataset,
        assembly=assembly,
        window_size=window_size,
        region=region,
        cell_line=cell_line
    )
    data_path = data_path_placeholder.format(root=root)
    label_path_placeholder = "{{root}}/{dataset}/{assembly}/{window_size}/{region}.bed.xz".format(
        root=root,
        dataset=dataset,
        assembly=assembly,
        window_size=window_size,
        region=region
    )
    label_path = label_path_placeholder.format(root=root)

    if not os.path.exists(data_path):
        download(
            url=data_path_placeholder.format(root=repository)+get_parameter,
            path=data_path
        )
    if not os.path.exists(label_path):
        download(
            url=label_path_placeholder.format(root=repository)+get_parameter,
            path=label_path
        )

    dtypes = {
        "chrom": "str",
        "chromStart": "int",
        "chromEnd": "int",
        "strand": "str"
    }

    X = pd.read_csv(
        data_path,
        index_col=[0, 1, 2, 3],
        header=[0, 1],
        low_memory=False,
        dtype=dtypes
    )

    X.index.rename(
        list(dtypes.keys()),
        inplace=True
    )

    X.columns.names = (None,)*len(X.columns.names)

    if X.columns.levels[1].size == 1 and drop_unique_group_by:
        X = X.droplevel(1, axis=1)

    y = pd.read_csv(
        label_path,
        index_col=[0, 1, 2, 3],
        sep="\t",
        dtype=dtypes
    ).astype(int)

    y = y[[cell_line.replace("-", "").upper()]]

    return X, y
